/*
 * Copyright © 2018 Red Hat, Inc
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *       Christian J. Kellner <christian@kellner.me>
 */

#include "config.h"

#include "bolt-unix.h"

gboolean
bolt_pid_is_alive (pid_t pid)
{
  gulong p = (gulong) pid;
  char path[256];

  if (pid != 0)
    g_snprintf (path, sizeof (path), "/proc/%lu/stat", p);
  else
    g_snprintf (path, sizeof (path), "/proc/self/stat");

  return g_file_test (path, G_FILE_TEST_EXISTS);
}
